#!/usr/local/bin/perl

# @author Dilip Antony Joseph (dilip@cs.berekeley.edu)

use strict;
require Utils;

use File::Basename;

# THis flag must be set to 0 if Keyword hunter is actually run.  If it is 1, we assume that the Keyword hunter output is already present.
my $gNoRunFlag = 0;


my $VERSION = "0.5f";

my $TEST_CASE_DIR="cases";
my $BINARY_NAME = "KeywordHunter";
my $DUMP_DIR="cases/";
my $ERROR_EXIT_CODE=256;
my $SUCCESS_EXIT_CODE=0;

if ($#ARGV + 1 < 4) {
  die "Usage: <NumLateDays, 0-3> <ExtraCredit> <TcpDumpPoints> <CleanCodePoints>\n";
}

my $gLateDays = shift(@ARGV);


my $gExtraCreditPoints = shift(@ARGV);
if ($gExtraCreditPoints > 15) {
  die "Invalid tcpdumpoints $gExtraCreditPoints\n";
}

my $gTcpDumpPoints = shift(@ARGV);
if ($gTcpDumpPoints > 5) {
  die "Invalid tcpdumpoints $gTcpDumpPoints\n";
}

my $gCleanCodePoints = shift (@ARGV);
if ($gCleanCodePoints > 5) {
  die "Invalid clean code points $gCleanCodePoints\n";
}

if ( ! -f $BINARY_NAME) {
  die "Unable to find the binary file $BINARY_NAME in the current directory.\n";
}

print "\n\n";
print "Project 2 Evaluator version $VERSION\n";

my $MaxPoints = 0;
my $TotalPoints = 0;

my $gMaxFoundLinePoints = 0;
my $gMaxValidLinkTraversalPoints = 0;
my $gMaxIgnoredLinkTraversalPoints = 0;
my $gMaxRobotsIgnoredLinkTraversalPoints = 0;
my $gMaxSavedOutputPoints = 0;
my $gMaxCmdLineArgsPoints = 0;

my $gMyFoundLinePoints = 0;
my $gMyValidLinkTraversalPoints = 0;
my $gMyIgnoredLinkTraversalPoints = 0;
my $gMyRobotsIgnoredLinkTraversalPoints = 0;
my $gMySavedOutputPoints = 0;
my $gMyCmdLineArgsPoints = 0;



my $lTmpTotalPoints;
my $lTmpMaxPoints;


($lTmpTotalPoints, $lTmpMaxPoints) = Test_CmdLineArgs();
$TotalPoints += $lTmpTotalPoints;
$MaxPoints += $lTmpMaxPoints;
$gMaxCmdLineArgsPoints = $lTmpMaxPoints;
$gMyCmdLineArgsPoints = $lTmpTotalPoints;

# Run the test cases
my @lCaseFiles = <$TEST_CASE_DIR/CASE*.txt>;
my $lCaseFile;
foreach $lCaseFile (@lCaseFiles) {
  ($lTmpTotalPoints, $lTmpMaxPoints) = RunCase($lCaseFile);
  $TotalPoints += $lTmpTotalPoints;
  $MaxPoints += $lTmpMaxPoints;
}

print "\n\n";
print "============================================================================================\n";
print "Report\n";
print "######\n\n";

#print "\nNoRunFlag = $gNoRunFlag\n";

print "CmdLineArgsPoints \t\t\t= $gMyCmdLineArgsPoints\t (out of $gMaxCmdLineArgsPoints)\n";
print "FoundLinePoints \t\t\t= $gMyFoundLinePoints\t (out of $gMaxFoundLinePoints)\n";
print "ValidLinkTraversalPoints \t\t= $gMyValidLinkTraversalPoints\t (out of $gMaxValidLinkTraversalPoints)\n";
print "IgnoredLinkTraversalPoints \t\t= $gMyIgnoredLinkTraversalPoints\t (out of $gMaxIgnoredLinkTraversalPoints)\n";
print "RobotsIgnoredLinkTraversalPoints \t= $gMyRobotsIgnoredLinkTraversalPoints\t (out of $gMaxRobotsIgnoredLinkTraversalPoints)\n";
print "SavedOutputPoints \t\t\t= $gMySavedOutputPoints\t (out of $gMaxSavedOutputPoints)\n";

print "Tcpdump Points \t\t\t\t= $gTcpDumpPoints\t (out of 5)\n";
print "Clean Code Points \t\t\t= $gCleanCodePoints\t (out of 5)\n";
print "Extra Credit\t\t\t\t= $gExtraCreditPoints\t (out of 15)\n";

$TotalPoints += $gTcpDumpPoints;
$TotalPoints += $gCleanCodePoints;

my $lGrandExtra = 0;

print "\n\n";
print "#######################################################\n";
print "Late by : $gLateDays day(s)\n";
my $lGrandTotal = 0;
my $lLateMsg = "";
if ($gLateDays == 0) {
  $lGrandTotal = $TotalPoints;
  $lGrandExtra = $gExtraCreditPoints;
} elsif ($gLateDays == 1) {
  $lLateMsg = "(90 % of $TotalPoints)";
  $lGrandTotal = 0.9 * $TotalPoints;
  $lGrandExtra = 0.9 * $gExtraCreditPoints;
} elsif ($gLateDays == 2) {
  $lLateMsg = "(80 % of $TotalPoints)";
  $lGrandTotal = 0.8 * $TotalPoints;
  $lGrandExtra = 0.8 * $gExtraCreditPoints;
} elsif ($gLateDays == 3) {
  $lLateMsg = "(60 % of $TotalPoints)";
  $lGrandTotal = 0.6 * $TotalPoints;
  $lGrandExtra = 0.6 * $gExtraCreditPoints;
} else {
  $lGrandTotal = 0;
  $lGrandExtra = 0;
}
$lGrandTotal = int ($lGrandTotal + 0.5);
$lGrandExtra = int ($lGrandExtra + 0.5);

print "Extra Credits                    =((($lGrandExtra)))\t (max 15) $lLateMsg\n";
print "Grand Total (no extra credits)   =[[[$lGrandTotal]]]\t (out of ".($MaxPoints + 5 +5).") $lLateMsg\n";
print "Grand Total (with extra credits) ={{{".($lGrandTotal + $lGrandExtra)."}}}\t (max ".($MaxPoints + 5 + 5 + 15).") $lLateMsg\n";
print "\n\n";

print "Notes:\n\n";






sub Test_CmdLineArgs() {
  my $lTestName = "CmdLineArgs";

  my $lDumpFile = "$DUMP_DIR/proj2_dump.txt";

  my $lMaxPoints = 3;
  my $lPoints = 0;

  `./$BINARY_NAME > $lDumpFile`;
  if ($? == $ERROR_EXIT_CODE) {
    $lPoints = $lPoints + 1;
  }
 
  system("./$BINARY_NAME http://danjo.cs.berkeley.edu/index.html > $lDumpFile");
  if ($? == $ERROR_EXIT_CODE) {
    $lPoints ++;
  }
 
  system("./$BINARY_NAME http://danjo.cs.berkeley.edu/ee122/testcases/final/cmdline/start.html CMD_LINE_FOUND > $lDumpFile");
  if ($? == $SUCCESS_EXIT_CODE) {
    $lPoints ++;
  } else {
    print ("Invalid return code : $?\n");
  }
 
  if ($lPoints > $lMaxPoints) {
    die "Max Points exceeded for $lTestName\n";
  }

  print "\$\$\$ TestResult: CmdLineArgs: $lPoints (out of $lMaxPoints)\n";

  return ($lPoints, $lMaxPoints);
 
}


sub RunCase() {
  

  my $lCaseFile = shift(@_);
  print "\n\n";
  print "-----------------------------------------------------------------------------\n";
  print ">>> Running Case : $lCaseFile\n";

  my $lPoints = 0;

  my $lCaseAnswerFile = "$lCaseFile.answer";
  if (!-f $lCaseAnswerFile) {
    die "Unable to find the answer file $lCaseAnswerFile.\n";
  }

  #Extract the details of the case.
  my %lCaseDetails = readConfigFile ($lCaseFile);
  my $lStartUrl = $lCaseDetails{"StartUrl"};
  my $lLevel = $lCaseDetails{"Level"};
  my $lSearchKeyword = $lCaseDetails{"SearchKeyword"};
  my $lDesc = $lCaseDetails{"Description"};
  my $lMaxPoints = $lCaseDetails{"MaxPoints"};
  my $lOutputDir = $lCaseDetails{"OutputDir"};
  my $lTestOutputDir = "";

  
  if (defined($lOutputDir)) {
    $lTestOutputDir = "$lOutputDir.out";
    if ($gNoRunFlag == 1) {
      if (!-d $lTestOutputDir) {
        die "NoRunFlag set.  But directory html output directory $lTestOutputDir not found.\n";
      }
    } else {
      if ( -d $lTestOutputDir) {
        `rm -rf $lTestOutputDir`;
      }
      mkdir ($lTestOutputDir) || die "Unable to create directory $lTestOutputDir\n";
    }
  } 

  my $lDoFoundLineTest = $lCaseDetails{"DoFoundLineTest"};
  my $lDoValidLinkTraversalTest = $lCaseDetails{"DoValidLinkTraversalTest"};
  my $lDoIgnoredLinkTraversalTest = $lCaseDetails{"DoIgnoredLinkTraversalTest"};
  my $lDoRobotsIgnoredLinkTraversalTest = $lCaseDetails{"DoRobotsIgnoredLinkTraversalTest"};

  
  if (  !defined($lStartUrl) ||
        !defined($lLevel) ||
        !defined($lSearchKeyword) ||
        !defined($lMaxPoints) 
     ) {
    die "Invalid case file $lCaseFile.\n";
  }
  
  if (defined($lDesc)) {
    print "Description: $lDesc\n\n";
  }
  my $lCaseTmpFile = "$lCaseFile.out";
  if ( $gNoRunFlag != 1) {
    if (-f $lCaseTmpFile) {
      unlink($lCaseTmpFile) || die "Unable to delete existing case output file $lCaseTmpFile\n";
    }
    my $lTmpCmd = "./$BINARY_NAME $lStartUrl $lSearchKeyword $lTestOutputDir > $lCaseTmpFile";
    `$lTmpCmd`;
    if ($? != $SUCCESS_EXIT_CODE) {
      print "UNEXPECTED_RETURN_VALUE: $lTmpCmd returned exit code $?, expected $SUCCESS_EXIT_CODE\n";
    }

  } else {
    if (!-f $lCaseTmpFile) {
      die "NoRunFlag set but output file $lCaseTmpFile not found.\n";
    }
  }



  if ($lDoFoundLineTest > 0) {
    my $lPoints_Found = $lDoFoundLineTest * $lMaxPoints;
    
    $gMaxFoundLinePoints += $lPoints_Found;
    if (CheckFoundLine($lCaseTmpFile, $lCaseAnswerFile) == 1) {
      $lPoints += $lPoints_Found;
      $gMyFoundLinePoints += $lPoints_Found;
    }
  }

  if ($lDoValidLinkTraversalTest > 0) {
    my $lPoints_ValidLinkTraversalTest = $lDoValidLinkTraversalTest * $lMaxPoints;
    
    $gMaxValidLinkTraversalPoints += $lPoints_ValidLinkTraversalTest;
    my $lTmpPoints = CheckValidLinkTraversal($lCaseTmpFile, $lCaseAnswerFile);
    
    $gMyValidLinkTraversalPoints += ($lPoints_ValidLinkTraversalTest * $lTmpPoints);
    $lPoints += $lPoints_ValidLinkTraversalTest * $lTmpPoints;
  }

  if ($lDoIgnoredLinkTraversalTest > 0) {
    my $lPoints_IgnoredLinkTraversalTest = $lDoIgnoredLinkTraversalTest * $lMaxPoints;
    
    $gMaxIgnoredLinkTraversalPoints += $lPoints_IgnoredLinkTraversalTest;
    my $lTmpPoints = CheckIgnoredLinkTraversal($lCaseTmpFile, $lCaseAnswerFile);
    $lPoints += $lPoints_IgnoredLinkTraversalTest * $lTmpPoints;
    $gMyIgnoredLinkTraversalPoints += $lPoints_IgnoredLinkTraversalTest * $lTmpPoints;
  }

  if ($lDoRobotsIgnoredLinkTraversalTest > 0) {
    my $lPoints_RobotsIgnoredLinkTraversalTest = $lDoRobotsIgnoredLinkTraversalTest * $lMaxPoints;
    
    $gMaxRobotsIgnoredLinkTraversalPoints += $lPoints_RobotsIgnoredLinkTraversalTest;
    my $lTmpPoints = CheckRobotsIgnoredLinkTraversal($lCaseTmpFile, $lCaseAnswerFile);
    $lPoints += $lPoints_RobotsIgnoredLinkTraversalTest * $lTmpPoints;
    $gMyRobotsIgnoredLinkTraversalPoints += $lPoints_RobotsIgnoredLinkTraversalTest * $lTmpPoints;
  }

  # Output director should always be tested alone!!
  if ($lOutputDir ne "") {
    my $lTestMaxPoints = 1 * $lMaxPoints;
    $gMaxSavedOutputPoints += $lTestMaxPoints;
    my $lTmpFraction = CheckSavedOutput($lTestOutputDir, $lOutputDir);
    print "****$lTmpFraction\n";
    $lPoints += $lTestMaxPoints * $lTmpFraction;
    $gMySavedOutputPoints += $lTestMaxPoints * $lTmpFraction;
    
  }

  if ($lPoints > $lMaxPoints + 0.01) {
    die "Max Points ($lMaxPoints) exceeded ($lPoints) for Case $lCaseFile\n";
  }

  print "\n\n";
  print "\$\$\$ TestResult: $lCaseFile: $lPoints (out of $lMaxPoints)\n";

  print "\n";
  print "Your output:\n";
  print "******************************\n";
  DumpFile ($lCaseTmpFile);
  print "******************************\n";
  print "\n\n"; 
  print "Expected output:\n";
  print "******************************\n";
  DumpFile($lCaseAnswerFile);
  print "******************************\n";

  print "\n\n";
  print ">>> Finished Running Case : $lCaseFile\n";
  print "-----------------------------------------------------------------------------\n";

  return ($lPoints, $lMaxPoints);
}

sub CheckSavedOutput() {
  my $lTestDir = shift(@_);
  my $lAnsDir = shift(@_);

  my $lTotalDiffs = 0;
  my $lCorrectDiffs = 0;

  my @lAnsOutputFiles = <$lAnsDir/*.html>;
  my $lAnsOutputFile;
  foreach $lAnsOutputFile (@lAnsOutputFiles) {
    my $lDiffCmd = "diff $lAnsOutputFile $lTestDir/".basename($lAnsOutputFile);
    `$lDiffCmd`;
    if ($? == 0) {
      $lCorrectDiffs++;
    } else {
      # Do an ignore whitespace diff for lower points
      my $lDiffCmdIgnoreWS = "diff -b -w $lAnsOutputFile $lTestDir/".basename($lAnsOutputFile);
      `$lDiffCmdIgnoreWS`;
      if ($? == 0) {
        print "Note: diff with ignore whitespace option invoked as normal diff found differences.  diff with ignore whitespace will fetch you only half the regular points\n";
        $lCorrectDiffs += 0.5;
      }
    }
    $lTotalDiffs++;
  }

  return ($lCorrectDiffs/$lTotalDiffs);
}

sub CheckValidLinkTraversal() {
  my $lTestFile = shift(@_);
  my $lAnsFile = shift(@_);

  open (TESTFD, "$lTestFile") ||die "Unable to open file $lTestFile\n";
  open (ANSFD, "$lAnsFile") ||die "Unable to open file $lAnsFile\n";

  my @lTestLines = <TESTFD>;

  my $lFollowedLinkCountAns = 0;
  my $lFollowedLinkCountTest = 0;

  while (<ANSFD>) {
    my $lCurrLine = $_;
    chomp($lCurrLine);

    if ($lCurrLine =~ /action=follow$/) {
      $lFollowedLinkCountAns += 1;
      my $lTmpTestLine;

      foreach $lTmpTestLine (@lTestLines) {
        chomp($lTmpTestLine);

        if ($lTmpTestLine =~ /$lCurrLine/) {
          $lFollowedLinkCountTest += 1;
          last;
        }
      }
    }

    if ($lCurrLine =~ /^PAGE:/) {
      $lFollowedLinkCountAns += 1;
      my $lTmpTestLine;
      foreach $lTmpTestLine (@lTestLines) {
      	chomp($lTmpTestLine);
        if ($lTmpTestLine =~ /$lCurrLine/) {
          $lFollowedLinkCountTest += 1;
          last;
        }
      }
    }

  }
  
  close (ANSFD);
  close (TESTFD);
  return $lFollowedLinkCountTest/$lFollowedLinkCountAns;

}

sub CheckIgnoredLinkTraversal() {
  my $lTestFile = shift(@_);
  my $lAnsFile = shift(@_);

  open (TESTFD, "$lTestFile") ||die "Unable to open file $lTestFile\n";
  open (ANSFD, "$lAnsFile") ||die "Unable to open file $lAnsFile\n";


  my %lIgnoredLineCountHashAns;
  my %lIgnoredLineCountHashTest;
  
  while (<ANSFD>) {
    my $lCurrLine = $_;
    chomp($lCurrLine);
    $lCurrLine = TrimWS($lCurrLine);

      if ($lCurrLine =~ /action=already_seen_ignore$/) {
        my $lTmpCount = $lIgnoredLineCountHashAns{$lCurrLine};
        if ($lTmpCount == undef) {
          $lIgnoredLineCountHashAns{$lCurrLine} = 1;
        } else {
          $lIgnoredLineCountHashAns{$lCurrLine} = $lTmpCount + 1;
        }
      }
  }
 
  while (<TESTFD>) {
    my $lCurrLine = $_;
        chomp($lCurrLine);
    $lCurrLine = TrimWS($lCurrLine);
      if ($lCurrLine =~ /action=already_seen_ignore$/) {
        my $lTmpCount = $lIgnoredLineCountHashTest{$lCurrLine};
        if ($lTmpCount == undef) {
          $lIgnoredLineCountHashTest{$lCurrLine} = 1;
        } else {
          $lIgnoredLineCountHashTest{$lCurrLine} = $lTmpCount + 1;
        }
      }
  }
 
  my $lTmpKey;
  my $lAnsCount = 0;
  my $lTestCount = 0;

  foreach $lTmpKey (keys %lIgnoredLineCountHashAns) {
    $lAnsCount ++;

    my $lTmp1 = $lIgnoredLineCountHashAns{$lTmpKey};
    my $lTmp2 = $lIgnoredLineCountHashTest{$lTmpKey};
#if ($lTmp1 == $lTmp2) {
      $lTestCount += $lTmp2/$lTmp1;
#    }
  }
  close (ANSFD);
  close (TESTFD);

  print "CHKKK>>>>>>>>$lTestCount/$lAnsCount\n";
  return $lTestCount/$lAnsCount;

}

sub CheckRobotsIgnoredLinkTraversal() {
  my $lTestFile = shift(@_);
  my $lAnsFile = shift(@_);

  open (TESTFD, "$lTestFile") ||die "Unable to open file $lTestFile\n";
  open (ANSFD, "$lAnsFile") ||die "Unable to open file $lAnsFile\n";


  my %lIgnoredLineCountHashAns;
  my %lIgnoredLineCountHashTest;
  
  while (<ANSFD>) {
    my $lCurrLine = $_;
    chomp($lCurrLine);
      if ($lCurrLine =~ /action=robots_ignore$/) {
        my $lTmpCount = $lIgnoredLineCountHashAns{$lCurrLine};
        if ($lTmpCount == undef) {
          $lIgnoredLineCountHashAns{$lCurrLine} = 1;
        } else {
          $lIgnoredLineCountHashAns{$lCurrLine} = $lTmpCount + 1;
        }
      }
  }
 
  while (<TESTFD>) {
    my $lCurrLine = $_;
      if ($lCurrLine =~ /action=robots_ignore$/) {
        chomp($lCurrLine);
        my $lTmpCount = $lIgnoredLineCountHashTest{$lCurrLine};
        if ($lTmpCount == undef) {
          $lIgnoredLineCountHashTest{$lCurrLine} = 1;
        } else {
          $lIgnoredLineCountHashTest{$lCurrLine} = $lTmpCount + 1;
        }
      }
  }
 
  my $lTmpKey;
  my $lAnsCount = 0;
  my $lTestCount = 0;

  foreach $lTmpKey (keys %lIgnoredLineCountHashAns) {
    $lAnsCount ++;

    my $lTmp1 = $lIgnoredLineCountHashAns{$lTmpKey};
    my $lTmp2 = $lIgnoredLineCountHashTest{$lTmpKey};
    if ($lTmp1 == $lTmp2) {
      $lTestCount++;
    }
  }
  close (ANSFD);
  close (TESTFD);
  
  print ">>>>>>>>$lTestCount/$lAnsCount\n";
  return $lTestCount/$lAnsCount;

}

sub CheckFoundLine() {
  my $lTestFile = shift(@_);
  my $lAnsFile = shift(@_);

  my $lAnsLine;

  my $lCorrectAnswer = 1; #found

  open (AnsFd, "$lAnsFile") || die "Unable to open $lAnsFile for reading.\n";
  while(<AnsFd>) {
    if (/^FOUND_KEYWORD:/) {
      $lAnsLine = $_;
      $lCorrectAnswer = 1;
      last;
    } elsif (/^NOT_FOUND_KEYWORD:/) {
      $lAnsLine = $_;
      $lCorrectAnswer = 0;
    }
  }
  if (!defined($lAnsLine)) {
    die "CheckFoundLine: Invalid answer file!\n";
  }
  close (AnsFd);

  my $lNOT_FOUNDFlag = 0;

  my $lFoundFlag = 0;
  open (TestFd, "$lTestFile") || die "Unable to open $lTestFile for reading.\n";
  while(<TestFd>) {
    if(/^$lAnsLine/) {
      $lFoundFlag = 1;
      #last;
    }
    if (/NOT_FOUND_KEYWORD:/) {
      $lNOT_FOUNDFlag = 1;
    }
  }
  close (TestFd);

  if ($lNOT_FOUNDFlag == 1 && $lCorrectAnswer == 1) {
    print(">>> Error: Correct answer is that the keyword was found.  But there is a NOT_FOUND_KEYWORD line. So test fails even irrespective of where FOUND_KEYWORD is present or not\n");
    return 0;
  }
  return $lFoundFlag;
  
}

sub DumpFile() {
  my $lFileToDump = shift(@_);
  open (DUMPFD, "$lFileToDump") || die "Unable to open file $lFileToDump for dumping\n";
  while(<DUMPFD>) {
    print $_;
  }
  close (DUMPFD);
}

sub TrimWS() {
  my $lStr = shift(@_);
  $lStr =~ s/^\s+//;
  $lStr =~ s/\s+$//;
  return $lStr;
}
